import pytest


def test_sum():
    assert 1 + 1 == 2


# @pytest.mark.xfail
def test_fail():
    assert 1 + 1 == 3


def test_dict_fail():
    assert {} == {"test_key": "test_value"}


@pytest.mark.custom1
def test_mark_custom():
    assert True


@pytest.mark.custom
@pytest.mark.custom_two
def test_mark_custom_two():
    assert True


@pytest.mark.skip
def test_mark_skip():
    assert 1 + 1 == 3


@pytest.mark.parametrize(
    "a, b, result",
    [
        (1, 1, 2),
        (1, 2, 3),
    ],
)
def test_mark(a, b, result):
    assert a + b == result


@pytest.mark.parametrize(
    "sequence",
    ["red", "*", "[+]", "[]", "/"],
)
def test_weird_fixtures(sequence):
    assert True



from textual.binding import Binding
from textual.message import Message
from textual.containers import Horizontal
from textual.reactive import reactive
from textual.widgets import Button, Rule

from rich.text import Text
from textual_slidecontainer import SlideContainer
from textual_tags import Tags, Tag, TagInput


class TreeFilter(SlideContainer):
    test_results_ready: reactive[bool] = reactive(False, init=False)
    markers: reactive[list[str]] = reactive([])

    def __init__(self, *args, **kwargs):
        super().__init__(
            slide_direction="down",
            floating=False,
            start_open=False,
            duration=0.5,
            *args,
            **kwargs,
        )

    def compose(self):
        yield MarkersFilter(tag_values=self.markers, show_x=True).data_bind(
            TreeFilter.markers
        )
        yield Rule()
        yield ResultFilter(id="horizontal_result_filter").data_bind(
            test_results_ready=TreeFilter.test_results_ready
        )


class ResultFilter(Horizontal):
    test_results_ready: reactive[bool] = reactive(False, init=False)

    def watch_test_results_ready(self):
        if not self.test_results_ready:
            self.border_title = Text.from_markup("No Tests have been run yet")
        else:
            self.border_title = Text.from_markup(
                ":magnifying_glass_tilted_right: Test Result Filter (click to toggle)"
            )
        display_style = "block" if self.test_results_ready else "none"
        self.query_children(".filter-button").set_styles(f"display:{display_style};")
        self.query_children(Button).last().display = not self.test_results_ready

    def compose(self):
        yield FilterButton(
            label=Text.from_markup("Marked: :star:"),
            id="button_filter_favourites",
            classes="filter-button",
        )
        yield FilterButton(
            label=Text.from_markup("Passed: :white_check_mark:"),
            id="button_filter_passed",
            classes="filter-button",
        )
        yield FilterButton(
            label=Text.from_markup("Failed: :x:"),
            id="button_filter_failed",
            classes="filter-button",
        )
        yield FilterButton(
            label=Text.from_markup("Skipped: [on yellow]:next_track_button: [/]"),
            id="button_filter_skipped",
            classes="filter-button",
        )
        yield Button(
            "Result Filter is available once tests are run",
            variant="primary",
            id="button_info",
        )


class FilterButton(Button):
    """Button for filtering the TestTree"""

    filter_is_active: reactive[bool] = reactive(True)

    def on_button_pressed(self):
        self.filter_is_active = not self.filter_is_active

    def watch_filter_is_active(self):
        self.variant = "success" if self.filter_is_active else "error"


class MarkersFilter(Tags):
    markers: reactive[list[str]] = reactive([])
    BINDINGS = [Binding("f", "mark_tests_with_markers", "⭐ Mark")]

    class Marked(Message):
        def __init__(self, current_tag: Tag) -> None:
            self.current_tag = current_tag
            super().__init__()

        @property
        def control(self):
            return self.current_tag

    async def watch_markers(self):
        self.query_one("#input_tag", TagInput).placeholder = "Select markers..."
        self.tag_values = set(self.markers)
        await self.query(Tag).remove()
        self.call_later(self._populate_with_tags)

    def action_mark_tests_with_markers(self):
        current_tag = self.app.focused.value
        self.post_message(self.Marked(current_tag=current_tag))

        # self.notify(f'unselected: {self.unselected_tags}')
        # self.selected_tags = set(self.markers)
        # await super().watch_tag_values()
        # super().watch_allow_new_tags()



from __future__ import annotations

# from collections.abc import Iterator
# import datetime

# import pathlib
import re

# from textwrap import dedent
# from textwrap import indent
# from typing import Any
from typing import TypedDict

# import packaging.version
import platformdirs
from requests_cache import CachedResponse
from requests_cache import CachedSession
from requests_cache import OriginalResponse
from requests_cache import SQLiteCache

DEVELOPMENT_STATUS_CLASSIFIERS = (
    "Development Status :: 1 - Planning",
    "Development Status :: 2 - Pre-Alpha",
    "Development Status :: 3 - Alpha",
    "Development Status :: 4 - Beta",
    "Development Status :: 5 - Production/Stable",
    "Development Status :: 6 - Mature",
    "Development Status :: 7 - Inactive",
)
ADDITIONAL_PROJECTS = {  # set of additional projects to consider as plugins
    "logassert",
    "logot",
    "nuts",
    "flask_fixture",
    "databricks-labs-pytester",
    "tursu",
}


def _escape_rst(text: str) -> str:
    """Rudimentary attempt to escape special RST characters to appear as
    plain text."""
    text = (
        text.replace("*", "\\*")
        .replace("<", "\\<")
        .replace(">", "\\>")
        .replace("`", "\\`")
    )
    text = re.sub(r"_\b", "", text)
    return text


def _project_response_with_refresh(
    session: CachedSession, name: str, last_serial: int
) -> OriginalResponse | CachedResponse:
    """Get a http cached pypi project

    force refresh in case of last serial mismatch
    """
    response = session.get(f"https://pypi.org/pypi/{name}/json")
    if int(response.headers.get("X-PyPI-Last-Serial", -1)) != last_serial:
        response = session.get(f"https://pypi.org/pypi/{name}/json", refresh=True)
    return response


def _get_session() -> CachedSession:
    """Configures the requests-cache session"""
    cache_path = platformdirs.user_cache_path("pytest-plugin-list")
    cache_path.mkdir(exist_ok=True, parents=True)
    cache_file = cache_path.joinpath("http_cache.sqlite3")
    return CachedSession(backend=SQLiteCache(cache_file))


def _pytest_plugin_projects_from_pypi(session: CachedSession) -> dict[str, int]:
    response = session.get(
        "https://pypi.org/simple",
        headers={"Accept": "application/vnd.pypi.simple.v1+json"},
        refresh=True,
    )
    return {
        name: p["_last-serial"]
        for p in response.json()["projects"]
        if (
            (name := p["name"]).startswith(("pytest-", "pytest_"))
            or name in ADDITIONAL_PROJECTS
        )
    }


class PluginInfo(TypedDict):
    """Relevant information about a plugin to generate the summary."""

    name: str
    summary: str
    last_release: str
    status: str
    requires: str


# async def get_plugin_list() -> dict[str, int]:
async def get_plugin_list() -> list[str]:
    session = _get_session()
    name_2_serial = _pytest_plugin_projects_from_pypi(session)
    return list(name_2_serial.keys())


# def iter_plugins() -> Iterator[PluginInfo]:
#     session = _get_session()
#     name_2_serial = _pytest_plugin_projects_from_pypi(session)
#
#
#     # for name, last_serial in tqdm(name_2_serial.items(), smoothing=0):
#     for name, last_serial in name_2_serial.items():
#         response = _project_response_with_refresh(session, name, last_serial)
#         if response.status_code == 404:
#             # Some packages, like pytest-azurepipelines42, are included in https://pypi.org/simple
#             # but return 404 on the JSON API. Skip.
#             continue
#         response.raise_for_status()
#         info = response.json()["info"]
#         if "Development Status :: 7 - Inactive" in info["classifiers"]:
#             continue
#         for classifier in DEVELOPMENT_STATUS_CLASSIFIERS:
#             if classifier in info["classifiers"]:
#                 status = classifier[22:]
#                 break
#         else:
#             status = "N/A"
#         requires = "N/A"
#         if info["requires_dist"]:
#             for requirement in info["requires_dist"]:
#                 if re.match(r"pytest(?![-.\w])", requirement):
#                     requires = requirement
#                     break
#
#         def version_sort_key(version_string: str) -> Any:
#             """
#             Return the sort key for the given version string
#             returned by the API.
#             """
#             try:
#                 return packaging.version.parse(version_string)
#             except packaging.version.InvalidVersion:
#                 # Use a hard-coded pre-release version.
#                 return packaging.version.Version("0.0.0alpha")
#
#         releases = response.json()["releases"]
#         for release in sorted(releases, key=version_sort_key, reverse=True):
#             if releases[release]:
#                 release_date = datetime.date.fromisoformat(
#                     releases[release][-1]["upload_time_iso_8601"].split("T")[0]
#                 )
#                 last_release = release_date.strftime("%b %d, %Y")
#                 break
#         name = f":pypi:`{info['name']}`"
#         summary = ""
#         if info["summary"]:
#             summary = _escape_rst(info["summary"].replace("\n", ""))
#         yield {
#             "name": name,
#             "summary": summary.strip(),
#             "last_release": last_release,
#             "status": status,
#             "requires": requires,
#         }



from types import FunctionType, NoneType
from typing import Any
from collections import defaultdict
import os
import shutil
import re
from enum import Enum
from pathlib import Path
import subprocess

import asyncio
from pytest import Config, OptionGroup
from pytest import Item, Class, Function
from _pytest.nodes import Node

from ayu.constants import WEB_SOCKET_PORT, WEB_SOCKET_HOST


class NodeType(str, Enum):
    DIR = "DIR"
    MODULE = "MODULE"
    CLASS = "CLASS"
    FUNCTION = "FUNCTION"
    COROUTINE = "COROUTINE"


class EventType(str, Enum):
    COLLECTION = "COLLECTION"
    SCHEDULED = "SCHEDULED"
    OUTCOME = "OUTCOME"
    REPORT = "REPORT"
    COVERAGE = "COVERAGE"
    PLUGIN = "PLUGIN"
    OPTIONS = "OPTIONS"
    DEBUG = "DEBUG"


class TestOutcome(str, Enum):
    PASSED = "PASSED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    QUEUED = "QUEUED"
    # handle those
    XFAILED = "XFAILED"
    XPASSED = "XPASSED"
    ERROR = "XPASSED"


class OptionType(str, Enum):
    INT = "INT"
    LIST = "LIST"
    STR = "STR"
    BOOL = "BOOL"
    SELECTION = "SELECTION"
    UNKNOWN = "UNKNOWN"


async def run_plugin_collection(command: str):
    """Collect All Tests without running them"""

    process = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    await process.wait()


async def run_test_collection(command: str):
    """Collect All Tests without running them"""

    process = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    await process.wait()


async def run_all_tests(command: str):
    """Run all selected tests"""

    return await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )


def get_nice_tooltip(node_data: dict) -> str | None:
    tooltip_str = ""
    # tooltip_str = f"{node_data['name'].replace("[", "\["):^20}\n"
    # tooltip_str += f"[red strike]{node_data['name'].replace('[', '\['):^20}[/]\n"
    #
    # status = node_data["status"].replace("[", "\[")
    # tooltip_str += f"\n[yellow]{status}[/]\n\n"
    return tooltip_str


def get_preview_test(file_path: str, start_line_no: int) -> str:
    """Read the test file from nodeid and use the linenumber
    and some rules to display the test function"""
    with open(Path(file_path), "r") as file:
        file_lines = file.readlines()
        last_line_is_blank = False
        end_line_no = None
        for line_no, line in enumerate(file_lines[start_line_no:], start=start_line_no):
            if not line.strip():
                last_line_is_blank = True
                continue
            if (
                line.strip().startswith(("def ", "class ", "async def ", "@"))
                and last_line_is_blank
            ):
                end_line_no = line_no - 1
                break
            last_line_is_blank = False
        return "".join(file_lines[start_line_no:end_line_no]).rstrip()


def get_ayu_websocket_host_port() -> tuple[str, int]:
    host: str = os.environ.get("AYU_HOST", WEB_SOCKET_HOST)
    port: int = int(os.environ.get("AYU_PORT", WEB_SOCKET_PORT))
    return host, port


def remove_ansi_escapes(string_to_remove: str) -> str:
    """Remove ansi escaped strings from colored pytest output"""
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", string_to_remove)


def uv_is_installed():
    if shutil.which("uv"):
        return True
    return False


def project_is_uv_managed():
    toml_path = Path.cwd() / "pyproject.toml"
    return toml_path.exists()


def ayu_is_run_as_tool():
    result = subprocess.run(
        "uv tree --package ayu".split(), capture_output=True, text=True
    )
    if result.stdout:
        return False
    return True


def build_dict_tree(items: list[Item]) -> dict:
    markers = set()

    def create_node(node: Node) -> dict[Any, Any]:
        markers.update([mark.name for mark in node.own_markers])
        return test_node_to_dict(node=node)

    def add_node(node_list: list[Node], sub_tree: dict):
        if not node_list:
            return

        # take root node
        current_node = node_list.pop(0)
        node_dict = create_node(node=current_node)

        existing_node = next(
            (
                node
                for node in sub_tree["children"]
                if node["nodeid"] == current_node.nodeid
            ),
            None,
        )

        if existing_node is None:
            sub_tree["children"].append(node_dict)
            existing_node = node_dict

        add_node(
            node_list=node_list,
            sub_tree=existing_node,
        )

    tree: dict[Any, Any] = {}
    root = items[0].listchain()[1]
    tree[root.name] = create_node(node=root)

    for item in items:
        # gets all parents except session
        parts_to_collect = item.listchain()[1:]
        add_node(node_list=parts_to_collect[1:], sub_tree=tree[root.name])

    return {"tree": tree, "meta": {"test_count": len(items), "markers": list(markers)}}


def get_coverage_data(coverage_file=".coverage"):
    import coverage

    cov = coverage.Coverage(data_file=coverage_file)
    cov.load()

    report_dict = {}

    all_files = cov.get_data().measured_files()
    for file_path in sorted(all_files):
        file_data = cov.analysis2(file_path)
        # analysis2 returns: (0:filename, 1:statements, 2:excluded, 3:missing, 4:partial)
        total_statements = len(file_data[1])  # All statements
        missing_statements = len(file_data[3])  # Uncovered statements
        coverage_percent = (
            (total_statements - missing_statements) / total_statements * 100
            if total_statements > 0
            else 0
        )

        # Store data for this file
        displayed_path = Path(file_path).relative_to(Path.cwd()).as_posix()
        report_dict[displayed_path] = {
            "n_statements": total_statements,
            "n_missed": missing_statements,
            "percent_covered": round(coverage_percent, 2),
            "lines_missing": file_data[3],  # List of uncovered line numbers
        }

    return report_dict


def test_node_to_dict(node: Node) -> dict[str, Any]:
    return {
        "name": node.name,
        "nodeid": node.nodeid,
        "markers": [mark.name for mark in node.own_markers],
        "path": node.path.as_posix(),
        "lineno": node.reportinfo()[1]
        if isinstance(node, Class)
        else (node.location[1] if isinstance(node, Function) else 0),
        "parent_name": node.parent.name if node.parent else None,
        "parent_type": type(node.parent).__name__.upper() if node.parent else None,
        "type": type(node).__name__.upper(),
        "favourite": False,
        "status": "",
        "children": [],
    }


def build_bar(percentage: float) -> str:
    SEGS = ["▉", "▊", "▋", "▌", "▍", "▎", "▏", ""]
    tens = int(percentage // 10)
    # can stay float cause we dont multiply it with strings
    rest = percentage - tens * 10
    not_tens = int((100 - percentage) // 10)

    tens_bar = f"[green on green]{tens * ' '}[/]"

    if rest == 0:
        rest_bar = ""
    elif rest < 2:
        rest_bar = f"[green on red]{SEGS[5]}[/]"
    elif rest < 4:
        rest_bar = f"[green on red]{SEGS[3]}[/]"
    elif rest < 6:
        rest_bar = f"[green on red]{SEGS[2]}[/]"
    elif rest < 8:
        rest_bar = f"[green on red]{SEGS[1]}[/]"
    elif rest < 10:
        rest_bar = f"[green on red]{SEGS[0]}[/]"

    not_tens_bar = f"[on red]{not_tens * ' '}[/]"
    return tens_bar + rest_bar + not_tens_bar


def build_plugin_dict(conf: Config) -> dict:
    # from pprint import pprint

    plugin_infos = sorted(
        set(
            (conf.pluginmanager.get_name(p), dist.version)
            for p, dist in conf.pluginmanager.list_plugin_distinfo()
        )
    )

    all_plugins_dict: dict[str, dict] = defaultdict(dict)
    for plugin_name, version in plugin_infos:
        clean_name = (
            plugin_name[7:] if plugin_name.startswith("pytest") else plugin_name
        )
        plugin_dict = {}
        plugin_group = conf._parser.getgroup(clean_name)
        group_name = plugin_group.name
        group_description = plugin_group.description
        group_options = plugin_group.options
        if not group_options:
            continue
        else:
            group_option_list = []
            for option in group_options:
                option_dict = get_plugin_option_dict(option=option)
                group_option_list.append(option_dict)

        plugin_dict["name"] = group_name
        plugin_dict["version"] = version
        plugin_dict["description"] = group_description
        plugin_dict["options"] = group_option_list

        all_plugins_dict[plugin_name] = plugin_dict

    # Remove Later
    # from pprint import pprint
    # pprint(all_plugins_dict['asyncio'], sort_dicts=False)
    return all_plugins_dict


def get_plugin_option_dict(option: OptionGroup) -> dict[str, Any]:
    option_names = option.names()
    option_attrs = option.attrs()
    option_default = option_attrs.get("default")
    option_help = option_attrs.get("help")
    # set type as 'None' if type is desclared as a function e.g. for validation
    option_type = infer_option_type(option_attributes=option_attrs)
    option_choices = option_attrs.get("choices")
    option_destination = option_attrs.get("dest")

    option_dict = {}

    # handle enums, cause they are not serializable
    if isinstance(option_default, Enum):
        option_default = option_default.value
        option_choices = [choice.value for choice in option_choices]

    option_dict["names"] = option_names
    option_dict["default"] = option_default
    option_dict["help"] = option_help
    option_dict["type"] = option_type
    option_dict["choices"] = option_choices
    option_dict["dest"] = option_destination
    # option_dict["attrs"] = option_attrs

    return option_dict


def infer_option_type(option_attributes: dict) -> OptionType:
    option_default = option_attributes.get("default")
    option_type = option_attributes.get(
        "type", type(option_default) if option_default is not None else None
    )

    if isinstance(option_type, (FunctionType, NoneType)):
        return OptionType.STR
    elif option_type is bool:
        return OptionType.BOOL
    elif option_attributes.get("choices"):
        return OptionType.SELECTION
    elif option_type is str:
        return OptionType.STR
    elif option_type is list:
        return OptionType.LIST
    elif option_type is int:
        return OptionType.INT
    else:
        return OptionType.UNKNOWN

    return option_type


def get_pytest_current_options(conf: Config) -> dict:
    def handle_enums(value: Enum | Any):
        if isinstance(value, Enum):
            return value.value
        return value

    return {option: handle_enums(value) for option, value in conf.option._get_kwargs()}



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ayu.app import AyuApp
from textual import work
from textual.reactive import reactive
from rich.style import Style
from textual.binding import Binding
from textual.widgets import Tree
from textual.widgets.tree import TreeNode, TreeDataType

from rich.text import Text, TextType

from ayu.utils import (
    EventType,
    NodeType,
    TestOutcome,
    get_nice_tooltip,
)
from ayu.constants import OUTCOME_SYMBOLS

TOGGLE_STYLE = Style.from_meta({"toggle": True})


class TestTree(Tree):
    app: "AyuApp"
    BINDINGS = [
        Binding("j,down", "cursor_down"),
        Binding("k,up", "cursor_up"),
        Binding("f", "mark_test_as_fav", "⭐ Mark"),
    ]
    show_root = False
    auto_expand = True
    guide_depth = 2

    counter_queued: reactive[int] = reactive(0)
    counter_passed: reactive[int] = reactive(0)
    counter_failed: reactive[int] = reactive(0)
    counter_skipped: reactive[int] = reactive(0)
    counter_marked: reactive[int] = reactive(0)

    filtered_data_test_tree: reactive[dict] = reactive({}, init=False)
    filtered_counter_total_tests: reactive[int] = reactive(0, init=False)
    filter: reactive[dict] = reactive(
        {
            "show_favourites": True,
            "show_failed": True,
            "show_skipped": True,
            "show_passed": True,
        },
    )

    def on_mount(self):
        self.app.dispatcher.register_handler(
            event_type=EventType.SCHEDULED,
            handler=lambda data: self.mark_tests_as_running(data),
        )
        self.app.dispatcher.register_handler(
            event_type=EventType.OUTCOME,
            handler=lambda data: self.update_test_outcome(data),
        )

        return super().on_mount()

    def watch_filter(self):
        if self.filtered_data_test_tree:
            self.build_tree()

    def watch_filtered_counter_total_tests(self):
        self.update_border_title()

    def watch_filtered_data_test_tree(self):
        if self.filtered_data_test_tree:
            self.build_tree()

    def watch_counter_queued(self):
        self.update_border_title()

    def watch_counter_passed(self):
        self.update_border_title()

    def watch_counter_failed(self):
        self.update_border_title()

    def watch_counter_skipped(self):
        self.update_border_title()

    def watch_counter_marked(self):
        self.update_border_title()
        self.app.refresh_bindings()

    def build_tree(self):
        self.clear()
        self.reset_status_counters()
        self.counter_marked = 0
        self.update_tree(tree_data=self.filtered_data_test_tree)

    def filter_tests(self, tests): ...

    def update_tree(self, *, tree_data: dict[Any, Any]):
        parent = self.root

        def add_children(child_list: list[dict[Any, Any]], parent_node: TreeNode):
            for child in child_list:
                if child["children"]:
                    if not self.filter["show_favourites"] and child["favourite"]:
                        continue
                    new_node = parent_node.add(
                        label=child["name"], data=child, expand=True
                    )
                    add_children(child_list=child["children"], parent_node=new_node)

                    # if all children were filtered out, remove this node
                    if not new_node.children:
                        new_node.remove()
                else:
                    # TODO Make this cleaner, also check for MODULES to be not displayed
                    if not self.filter["show_favourites"] and child["favourite"]:
                        self.filtered_counter_total_tests -= 1
                        continue
                    if not self.filter["show_passed"] and (
                        child["status"] == TestOutcome.PASSED
                    ):
                        self.filtered_counter_total_tests -= 1
                        continue
                    if not self.filter["show_skipped"] and (
                        child["status"] == TestOutcome.SKIPPED
                    ):
                        self.filtered_counter_total_tests -= 1
                        continue
                    if not self.filter["show_failed"] and (
                        child["status"] == TestOutcome.FAILED
                    ):
                        self.filtered_counter_total_tests -= 1
                        continue

                    parent_node.add_leaf(label=child["name"], data=child)

                    if child["favourite"]:
                        self.counter_marked += 1

                    match child["status"]:
                        case TestOutcome.PASSED:
                            self.counter_passed += 1
                        case TestOutcome.SKIPPED:
                            self.counter_skipped += 1
                        case TestOutcome.FAILED:
                            self.counter_failed += 1

        for key, value in tree_data.items():
            if isinstance(value, dict) and "children" in value and value["children"]:
                node: TreeNode = parent.add(key, data=value)
                node.expand()
                add_children(value["children"], node)
            else:
                parent.add_leaf(key, data=key)

        # set initial cursor line
        self.cursor_line: int = 0 if self.cursor_line < 0 else self.cursor_line

        # TODO remove empty nodes after filtering
        # for node in self._tree_nodes.values():
        #     self.log(f'{node}')
        #     if node:
        #         if node.children:
        #             continue
        #         else:
        #             node.remove()

    def update_test_outcome(self, test_result: dict):
        for node in self._tree_nodes.values():
            if node.data and (node.data["nodeid"] == test_result["nodeid"]):
                outcome = test_result["outcome"]
                node.data["status"] = outcome
                node.refresh()
                node.parent.refresh()
                self.counter_queued -= 1
                match outcome:
                    case TestOutcome.PASSED:
                        self.counter_passed += 1
                    case TestOutcome.FAILED:
                        self.counter_failed += 1
                    case TestOutcome.SKIPPED:
                        self.counter_skipped += 1

                self.update_collapse_state_on_test_run(node=node)
                self.update_filtered_data_test_tree(
                    nodeid=node.data["nodeid"], new_status=outcome
                )

    def update_collapse_state_on_test_run(self, node: TreeNode):
        def all_child_tests_passed(parent: TreeNode):
            return all(
                [
                    all_child_tests_passed(parent=child)
                    if child.data["type"] == NodeType.CLASS
                    else child.data["status"] in [TestOutcome.PASSED]
                    for child in parent.children
                ]
            )

        if node.parent.data["type"] == NodeType.CLASS:
            self.update_collapse_state_on_test_run(node=node.parent)
        if all_child_tests_passed(parent=node.parent):
            node.parent.collapse()

    def reset_status_counters(self) -> None:
        self.counter_queued = 0
        self.counter_passed = 0
        self.counter_skipped = 0
        self.counter_failed = 0
        self.filtered_counter_total_tests = self.app.counter_total_tests

    def mark_tests_as_running(self, nodeids: list[str]) -> None:
        self.root.expand_all()
        self.reset_status_counters()
        for node in self._tree_nodes.values():
            if node.data and (node.data["nodeid"] in nodeids):
                node.data["status"] = TestOutcome.QUEUED
                self.counter_queued += 1

    def on_tree_node_selected(self, event: Tree.NodeSelected):
        # self.notify(f"{','.join(event.node.data['markers'])}")
        ...
        # Run Test

    def mark_test_as_fav_from_markers(self, marker: str):
        for node in self._tree_nodes.values():
            if node.data and marker in node.data["markers"]:
                self.action_mark_test_as_fav(node=node)

    def action_mark_test_as_fav_from_search(self, nodeid: str):
        node_to_mark = self.get_node_by_nodeid(nodeid=nodeid)
        self.action_mark_test_as_fav(node=node_to_mark)

    @work(thread=True)
    def action_mark_test_as_fav(
        self, node: TreeNode | None = None, parent_val: bool | None = None
    ):
        # if no node given, select node under cursor
        if node is None:
            node = self.cursor_node

        if parent_val is None:
            parent_val = not node.data["favourite"]

        # mark all childs the same as parent
        if node.children:
            node.data["favourite"] = parent_val
            node.refresh()
            # self.update_filtered_data_test_tree(
            #     nodeid=node.data["nodeid"],
            #     is_fav=parent_val,
            # )
            for child in node.children:
                self.action_mark_test_as_fav(node=child, parent_val=parent_val)
        else:
            if node.data["favourite"] != parent_val:
                self.counter_marked += 1 if parent_val else -1
            node.data["favourite"] = parent_val
            node.refresh()
            # self.update_filtered_data_test_tree(
            #     nodeid=node.data["nodeid"],
            #     is_fav=parent_val,
            # )
            # self.mutate_reactive(TestTree.filtered_data_test_tree)

        # Unfavourite all parents, if a single child not is not favourited
        if not node.data["favourite"]:
            parent_node = node.parent
            while parent_node.data is not None:
                parent_node.data["favourite"] = node.data["favourite"]
                parent_node.refresh()
                parent_node = parent_node.parent

    def update_filtered_data_test_tree(
        self,
        nodeid: str,
        is_fav: bool | None = None,
        new_status: TestOutcome | None = None,
    ):
        def update_filtered_node(child_list: list):
            for child in child_list:
                if child["nodeid"] == nodeid:
                    if is_fav is not None:
                        child["favourite"] = is_fav
                    if new_status:
                        child["status"] = new_status
                    return True
                if child["children"]:
                    update_filtered_node(child_list=child["children"])

        for key, val in self.filtered_data_test_tree.items():
            if val["nodeid"] == nodeid:
                if is_fav is not None:
                    val["favourite"] = is_fav
                if new_status:
                    val["status"] = new_status
                return True
            if val["children"]:
                update_filtered_node(child_list=val["children"])

    def process_label(self, label: TextType) -> Text:
        """Subclassed to handle [/] sequences, e.g. in parametrized tests"""
        text_label = label
        first_line = text_label.split()[0]
        return first_line

    def render_label(
        self, node: TreeNode[TreeDataType], base_style: Style, style: Style
    ) -> Text:
        fav_substring = Text.from_markup("⭐ ") if node.data["favourite"] else Text()
        escaped_name_substring = Text(node.data["name"], style)
        # Render Classes, Modules and Folders
        if node._allow_expand:
            prefix = (
                self.ICON_NODE_EXPANDED if node.is_expanded else self.ICON_NODE,
                base_style + TOGGLE_STYLE,
            )
            amount_test_results = self.get_number_of_tests_queued_of_node(node=node)
            amount_tests_passed = self.get_number_of_passed_tests_of_node(node=node)
            if (
                node.data["type"] in [NodeType.CLASS, NodeType.MODULE]
                and amount_test_results
            ):
                run_status_label = Text(f" {amount_tests_passed}/{amount_test_results}")
                run_label_color = (
                    "green" if amount_tests_passed == amount_test_results else "red"
                )
            else:
                run_status_label = ""
                run_label_color = None

            module_label = Text.assemble(
                prefix,
                fav_substring,
                escaped_name_substring,
                run_status_label,
                style=Style(color=run_label_color),
            )
            module_label.stylize(style)
            return module_label

        # Render Test Labels
        status_substring = Text.from_markup(
            f" {OUTCOME_SYMBOLS[node.data['status']]}" if node.data["status"] else ""
        )
        test_label = Text.assemble(
            fav_substring, escaped_name_substring, status_substring
        )
        return test_label

    def get_number_of_tests_queued_of_node(self, node: TreeNode) -> int:
        return len(
            [
                child
                for child in node.children
                if (child.data["type"] in [NodeType.FUNCTION, NodeType.COROUTINE])
                if child.data["status"]
            ]
        )

    def get_number_of_passed_tests_of_node(self, node: TreeNode) -> int:
        return len(
            [
                child
                for child in node.children
                if child.data["status"] == TestOutcome.PASSED
            ]
        )

    def on_mouse_move(self):
        return
        if self.hover_line != -1:
            data = self._tree_lines[self.hover_line].node.data
            self.tooltip = get_nice_tooltip(node_data=data)

    def highlight_marker_rows(self, marker: str):
        for node in self._tree_nodes.values():
            node._hover = False
            if node.data and (marker in node.data["markers"]):
                node._hover = True
                last_true_line = node.line
                self.refresh()
        self.hover_line = last_true_line

    def update_border_title(self):
        symbol = "hourglass_not_done" if self.counter_queued > 0 else "hourglass_done"
        tests_to_run = (
            self.filtered_counter_total_tests
            if not self.counter_marked
            else f":star: {self.counter_marked}/{self.filtered_counter_total_tests}"
        )

        self.border_title = Text.from_markup(
            f" :{symbol}: {self.counter_queued} | :x: {self.counter_failed}"
            + f" | :white_check_mark: {self.counter_passed} | :next_track_button: {self.counter_skipped}"
            + f" | Tests to run {tests_to_run} "
        )

    def get_node_by_nodeid(self, nodeid: str) -> TreeNode | None:
        for node in self._tree_nodes.values():
            if node.data and (node.data["nodeid"] == nodeid):
                return node
        return None

    @property
    def marked_tests(self):
        # TODO based on self.filtered_data_test_tree,
        # to run tests accordingly when filter is active
        marked_tests = []
        for node in self._tree_nodes.values():
            if (
                node.data
                and (node.data["type"] in [NodeType.FUNCTION, NodeType.COROUTINE])
                and node.data["favourite"]
            ):
                marked_tests.append(node.data["nodeid"])
        return marked_tests

    @property
    def test_nodes(self) -> list[TreeNode]:
        test_nodes = []
        for node in self._tree_nodes.values():
            if node.data:
                test_nodes.append(node)
        return test_nodes

    def reset_test_results(self):
        # reset self.filtered_data_test_tree,
        # to also reset results that were hidden by the filter
        self.reset_status_counters()
        for node in self._tree_nodes.values():
            if node.data:
                node.data["status"] = ""
                node.refresh()



import pytest
from pathlib import Path
from ayu.command_builder import Plugin, build_command

not_installed_cov_plugin = Plugin(
    name="pytest-cov",
    version="0.0.1",
    is_installed=False,
    options=["--cov src/ayu", "--report term-missing"],
)


@pytest.mark.parametrize(
    "plugins,tests_to_run,pytest_options,expected_command",
    (
        [None, "", None, "uv run pytest"],
        [
            [not_installed_cov_plugin],
            "",
            None,
            "uv run --with pytest-cov pytest --cov src/ayu --report term-missing",
        ],
        [
            None,
            Path("tests/test_cases"),
            None,
            "uv run pytest tests/test_cases",
        ],
        [None, "", ["--co", "--color=yes"], "uv run pytest --co --color=yes"],
    ),
)
def test_build_command(plugins, tests_to_run, pytest_options, expected_command):
    command = build_command(
        plugins=plugins,
        pytest_options=pytest_options,
        tests_to_run=tests_to_run,
    )
    assert command == expected_command



from pathlib import Path
import os

import pytest


@pytest.fixture()
def testcase_path() -> Path:
    return Path("tests/test_cases")


@pytest.fixture()
def test_host() -> str:
    os.environ["AYU_HOST"] = "localhost"
    return "localhost"


@pytest.fixture()
def test_port() -> int:
    os.environ["AYU_PORT"] = "1338"
    return 1338



from typing import Callable
from collections import defaultdict

from websockets.asyncio.server import serve
from websockets.asyncio.client import connect
from websockets.exceptions import ConnectionClosedOK, WebSocketException
import asyncio

from ayu.constants import MAX_EVENT_SIZE, WEB_SOCKET_HOST, WEB_SOCKET_PORT
from ayu.classes.event import Event
from ayu.utils import EventType, get_ayu_websocket_host_port


class EventDispatcher:
    def __init__(self, host: str = WEB_SOCKET_HOST, port: int = WEB_SOCKET_PORT):
        self.host = host
        self.port = port
        self.running = False
        self.server = None
        self.event_handler: defaultdict[EventType | None, list] = defaultdict(list)

        self.data = ""

    # Handler
    async def handler(self, websocket):
        while True:
            try:
                msg = await websocket.recv()
            except ConnectionClosedOK:
                break

            if self.event_handler:
                event = Event.deserialize(msg)
                event_type = event.event_type
                event_payload = event.event_payload
                match event_type:
                    case EventType.COLLECTION:
                        handlers = self.event_handler[EventType.COLLECTION]
                    case EventType.OUTCOME:
                        handlers = self.event_handler[EventType.OUTCOME]
                    case EventType.REPORT:
                        handlers = self.event_handler[EventType.REPORT]
                    case EventType.SCHEDULED:
                        handlers = self.event_handler[EventType.SCHEDULED]
                    case EventType.COVERAGE:
                        handlers = self.event_handler[EventType.COVERAGE]
                    case EventType.PLUGIN:
                        handlers = self.event_handler[EventType.PLUGIN]
                    case EventType.OPTIONS:
                        handlers = self.event_handler[EventType.OPTIONS]
                    case EventType.DEBUG:
                        handlers = self.event_handler[EventType.DEBUG]

            if handlers:
                for handler in handlers:
                    handler(event_payload)

            self.data = msg

    def register_handler(self, event_type: EventType, handler: Callable):
        self.event_handler[event_type].append(handler)

    def unregister_handler(self, event_type: EventType):
        # with asyncio.Lock():
        self.event_handler.pop(event_type)

    # Start Websocket Server
    async def start(self):
        await self.start_socket_server()

    # Stop Websocket Server
    async def stop(self):
        await asyncio.get_running_loop().create_future()

    async def start_socket_server(self):
        self.server = await serve(
            self.handler, self.host, self.port, max_size=MAX_EVENT_SIZE
        )
        await self.server.wait_closed()

    def get_data(self):
        return self.data


# send events
async def send_event(
    event: Event, host: str = WEB_SOCKET_HOST, port: int = WEB_SOCKET_PORT
):
    # host, port = get_ayu_websocket_host_port()
    uri = f"ws://{host}:{port}"

    async with connect(uri) as websocket:
        await websocket.send(message=event.serialize())


async def is_websocket_connected():
    host, port = get_ayu_websocket_host_port()
    uri = f"ws://{host}:{port}"
    try:
        async with connect(uri) as _websocket:
            return True
    except (WebSocketException, ConnectionRefusedError, OSError):
        return False


def check_connection():
    return asyncio.run(is_websocket_connected())



import os
from pathlib import Path
from platformdirs import user_data_dir

WEB_SOCKET_HOST = os.environ.get("AYU_HOST") or "localhost"
WEB_SOCKET_PORT = int(os.environ.get("AYU_PORT", 0)) or 1337
MAX_EVENT_SIZE = 2**30
# WEB_SOCKET_HOST = "localhost"
# WEB_SOCKET_PORT = 1337

OUTCOME_SYMBOLS = {
    "PASSED": ":white_check_mark:",
    "FAILED": ":cross_mark:",
    "XFAILED": "[on green]:cross_mark:[/]",
    "SKIPPED": "[on yellow]:next_track_button: [/]",
    "QUEUED": ":hourglass_not_done:",
}

# Some Plugin Options are not really parseable
# with the given approach, e.g. --cov-reset
# which dest is cov_source
# Will look later into that, and disable it for now

OPTIONS_TO_DISABLE = ["--cov-reset"]

PLUGIN_JSON_PATH = Path(
    user_data_dir(appname="ayu", appauthor=False, ensure_exists=True)
)
PLUGIN_JSON_FILE = PLUGIN_JSON_PATH.joinpath("pytest_plugins.json")



from pathlib import Path
import click

from ayu.app import AyuApp
from ayu.utils import uv_is_installed, project_is_uv_managed, ayu_is_run_as_tool


@click.group(
    context_settings={"ignore_unknown_options": True}, invoke_without_command=True
)
@click.version_option(prog_name="ayu")
@click.pass_context
@click.argument(
    "tests_path",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    required=False,
)
def cli(ctx, tests_path):
    if ayu_is_run_as_tool():
        print("ayu as tool")
    else:
        print("ayu as dependency")
    # return
    if not uv_is_installed():
        print("uv_not_installed, please install uv to use ayu")
        return
    if not project_is_uv_managed():
        print("Your project is not a valid python project")
        print("no pyproject.toml file detected")
        return

    if tests_path:
        app = AyuApp(test_path=tests_path)
        app.run()
    elif ctx.invoked_subcommand is None:
        app = AyuApp()
        app.run()
    else:
        pass


if __name__ == "__main__":
    cli()



from typing import TYPE_CHECKING, Any
import json


if TYPE_CHECKING:
    from ayu.app import AyuApp

from textual import on, work
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.binding import Binding
from textual.widgets import (
    Button,
    Footer,
    Label,
    Switch,
    Select,
    Input,
    Collapsible,
    Rule,
    DataTable,
)
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual_autocomplete import AutoComplete, TargetState, DropdownItem

from ayu.utils import OptionType, run_plugin_collection
from ayu.constants import OPTIONS_TO_DISABLE, PLUGIN_JSON_FILE, PLUGIN_JSON_PATH
from ayu.plugin_list_fetcher import get_plugin_list
from ayu.command_builder import Plugin, build_command


class ModalPlugin(ModalScreen):
    app: "AyuApp"
    available_plugin_list: list[str] = reactive([], init=False)
    """Plugins loaded from PyPi"""
    plugin_option_dict: reactive[dict] = reactive({}, init=False)
    """Dict of available plugins {plugin_name: plugin_dict}"""
    selected_options_dict: reactive[dict] = reactive({}, init=False)
    """Dict of active options {optionname:value}"""

    BINDINGS = [
        Binding("escape", "app.pop_screen", "Close", show=True),
    ]

    # @work(thread=True, description="Collect Plugins")
    async def on_mount(self):
        # Load Plugins on first Load
        if not PLUGIN_JSON_FILE.exists():
            self.update_plugins()
        else:
            self.load_plugins()

    def compose(self):
        with Vertical():
            yield Footer()

            # self.button.loading = True
            yield Button("Refresh plugin list", id="button_refresh_plugin")
            yield PluginInput(id="input_plugin_list")
            yield PluginAutoComplete(
                target="#input_plugin_list",
            ).data_bind(available_plugin_list=ModalPlugin.available_plugin_list)

            with VerticalScroll():
                for plugin, plugin_dict in self.app.plugin_option_dict.items():
                    yield PluginEntry(
                        plugin_name=plugin, plugin_dict=plugin_dict, installed=True
                    )

    def watch_plugin_option_dict(self):
        # self.notify(f"modal: {self.plugin_option_dict.keys()}", markup=False)
        # Refresh like this turns all other settings to default
        # self.refresh(recompose=True)
        for plugin, plugin_dict in self.plugin_option_dict.items():
            if plugin not in self.loaded_plugins:
                vs = self.query_one(VerticalScroll)
                vs.mount(PluginEntry(plugin_name=plugin, plugin_dict=plugin_dict))

    def watch_available_plugin_list(self):
        if self.available_plugin_list:
            # self.notify(f"{len(self.available_plugin_list)}", markup=False)
            self.query_one("#input_plugin_list", Input).display = True
            self.query_one(
                "#input_plugin_list", Input
            ).placeholder = (
                f"Type to search for plugins (found {len(self.available_plugin_list)})"
            )
            self.query_one("#button_refresh_plugin", Button).loading = False
        else:
            self.query_one("#input_plugin_list", Input).display = False
            self.query_one("#button_refresh_plugin", Button).loading = True

    def watch_selected_options_dict(self): ...

    @on(Button.Pressed, "#button_refresh_plugin")
    def refresh_plugin_list(self, event: Button.Pressed):
        # TODO Make Loading work on refresh
        self.update_plugins()

    @work(thread=True, description="Refresh Plugins")
    async def update_plugins(self):
        # Create Pluginfolder and load plugin_dict
        PLUGIN_JSON_PATH.mkdir(exist_ok=True, parents=True)
        PLUGIN_JSON_FILE.touch(exist_ok=True)
        await self.fetch_plugin_list()
        with open(PLUGIN_JSON_FILE, "w") as json_file:
            json.dump(self.available_plugin_list, json_file)

    def load_plugins(self):
        """load preloaded pluginlist from file"""
        with open(PLUGIN_JSON_FILE, "r") as json_file:
            self.available_plugin_list = json.load(json_file)

    # @work(thread=True)
    async def fetch_plugin_list(self):
        self.available_plugin_list = await get_plugin_list()
        self.notify(f"found {len(self.available_plugin_list)} plugins", timeout=1)

    @on(Input.Submitted, "#input_plugin_list")
    @work(thread=True, description="Load new PluginInfos")
    async def load_plugin_into_options(self, event: Input.Submitted):
        new_plugin_name = event.input.value
        if new_plugin_name not in self.available_plugin_list:
            self.notify(
                title="Error", message="plugin not in pluginlist", severity="warning"
            )
        else:
            new_plugin = Plugin(name=new_plugin_name, is_installed=False, options=[])
            # self.notify(f"{new_plugin}", markup=False)

            # Update option dict with new plugins
            command = build_command(plugins=[new_plugin], pytest_options=["--help"])
            await run_plugin_collection(command=command)
            self.query_one(Input).clear()
            # self.mutate_reactive(ModalPlugin.plugin_option_dict)

    @property
    def loaded_plugins(self) -> list[str]:
        return [plugin.plugin_name for plugin in self.query(PluginEntry)]


class PluginEntry(Vertical):
    def __init__(
        self,
        plugin_name: str,
        plugin_dict: dict,
        installed: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.plugin_name = plugin_name
        self.installed = installed
        self.plugin_dict = plugin_dict
        self.plugin = Plugin(
            name=self.plugin_name, is_installed=self.installed, options=[]
        )

    def compose(self):
        with PlugInCollapsible(title=self.plugin_name):
            for option_dict in self.plugin_dict["options"]:
                option_name = " ".join(option_dict["names"])

                # Skip certain options for now
                if option_name in OPTIONS_TO_DISABLE:
                    continue

                match option_dict["type"]:
                    case OptionType.BOOL:
                        yield BoolOption(option_dict=option_dict)
                    case OptionType.STR:
                        yield StringOption(option_dict=option_dict)
                    case OptionType.LIST:
                        yield ListOption(option_dict=option_dict)
                    case OptionType.SELECTION:
                        yield SelectionOption(option_dict=option_dict)


class PluginInput(Input):
    def on_mount(self):
        self.display = False


class PluginAutoComplete(AutoComplete):
    available_plugin_list: reactive[list[str]] = reactive([])

    def get_candidates(self, target_state: TargetState) -> list[DropdownItem]:
        # Filter candidates based on target_state.text
        return [DropdownItem(main=plugin) for plugin in self.available_plugin_list]

    # Display Infos on completion about plugin
    # Maybe even better on typing
    def post_completion(self) -> None:
        self.notify(f"press enter to add plugin {self.target.value}")
        return super().post_completion()

    # TODO Add ctrl + j/k navigation


class PlugInCollapsible(Collapsible):
    amount_changed: reactive[int] = reactive(0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.plugin = self.title

    def on_mount(self):
        self.update_amount()

    @on(Switch.Changed)
    @on(Input.Changed)
    @on(Select.Changed)
    @on(DataTable.RowSelected)
    def update_amount(self):
        self.amount_changed = sum(
            [widget.was_changed for widget in self._contents_list if widget.was_changed]
        )

    def watch_amount_changed(self):
        if self.amount_changed > 0:
            amount_str = (
                f"[$success]{self.amount_changed}/{len(self._contents_list)}[/]"
            )
        else:
            amount_str = f"[$error]{self.amount_changed}/{len(self._contents_list)}[/]"
        self.title = f"{amount_str} {self.plugin}"


class BoolOption(Vertical):
    app: "AyuApp"
    option: reactive[str] = reactive("", init=False)
    option_value: reactive[str] = reactive("", init=False)
    complete_option: reactive[str | None] = reactive(None, init=False)
    was_changed: reactive[bool] = reactive(False, init=False)

    def __init__(self, option_dict: dict[str, Any], *args, **kwargs) -> None:
        self.option_dict = option_dict
        super().__init__(*args, **kwargs)
        self.classes = "optionwidget"
        self.option = " ".join(option_dict["names"])
        self.option_value = option_dict["default"]

    def on_mount(self):
        self.query_one(Label).tooltip = self.option_dict["help"]
        # Set to current Value
        dest = self.option_dict["dest"]
        if dest != self.option_dict["default"]:
            self.query_one(Switch).value = self.app.selected_options_dict[dest]

    def compose(self):
        with Horizontal():
            yield Label(f"{self.option} [gray]{self.option_dict['dest']}[/]")
            yield Switch(value=self.option_dict["default"])
        yield Rule()
        return super().compose()

    def on_switch_changed(self, event: Switch.Changed):
        self.option_value = event.switch.value

    def watch_option_value(self):
        if self.option_value == self.option_dict["default"]:
            self.complete_option = None
        else:
            self.complete_option = f"{self.option}"

    def watch_was_changed(self):
        if self.was_changed:
            self.query_one(Label).update(
                f"[$success]{self.option}[/] [gray]{self.option_dict['dest']}[/]"
            )
        else:
            self.query_one(Label).update(
                f"{self.option} [gray]{self.option_dict['dest']}[/]"
            )

    # complete option string for the command builder
    def watch_complete_option(self):
        if self.complete_option is None:
            # self.app.options.pop(self.flag)
            self.was_changed = False
        else:
            # self.app.options[self.flag] = self.complete_flag
            self.was_changed = True

        # self.app.update_options()


class StringOption(Vertical):
    app: "AyuApp"
    option: reactive[str] = reactive("", init=False)
    option_value: reactive[str] = reactive("", init=False)
    complete_option: reactive[str | None] = reactive(None, init=False)
    was_changed: reactive[bool] = reactive(False, init=False)

    def __init__(self, option_dict: dict[str, str], *args, **kwargs) -> None:
        self.option_dict = option_dict
        super().__init__(*args, **kwargs)
        self.classes = "optionwidget"
        self.option = " ".join(option_dict["names"])
        self.option_value = option_dict["default"]

    def on_mount(self):
        self.query_one(Label).tooltip = self.option_dict["help"]
        # Set to current Value
        dest = self.option_dict["dest"]
        if dest != self.option_dict["default"]:
            value = self.app.selected_options_dict[dest]
            self.query_one(Input).value = f"{value}" if isinstance(value, list) else ""

    def compose(self):
        with Horizontal():
            yield Label(f"{self.option} [gray]{self.option_dict['dest']}[/]")
            yield Input(placeholder=f"default: {self.option_dict['default']}")
        yield Rule()
        return super().compose()

    def on_input_changed(self, event: Input.Changed):
        self.option_value = event.input.value

    def watch_option_value(self):
        if self.option_value in [self.option_dict["default"], ""]:
            self.complete_option = None
        else:
            self.complete_option = f"{self.option}={self.option_value}"

    def watch_was_changed(self):
        if self.was_changed:
            self.query_one(Label).update(
                f"[$success]{self.option}[/] [gray]{self.option_dict['dest']}[/]"
            )
        else:
            self.query_one(Label).update(
                f"{self.option} [gray]{self.option_dict['dest']}[/]"
            )

    def watch_complete_option(self):
        if self.complete_option is None:
            # self.app.options.pop(self.option)
            self.was_changed = False
        else:
            # self.app.options[self.option] = self.complete_option
            self.was_changed = True
        # self.app.update_options()


class SelectionOption(Vertical):
    app: "AyuApp"
    option: reactive[str] = reactive("", init=False)
    option_value: reactive[str] = reactive("", init=False)
    complete_option: reactive[str | None] = reactive(None, init=False)
    was_changed: reactive[bool] = reactive(False, init=False)

    def __init__(self, option_dict: dict[str, str | list], *args, **kwargs) -> None:
        self.option_dict = option_dict
        super().__init__(*args, **kwargs)
        self.classes = "optionwidget"
        self.option = " ".join(option_dict["names"])
        self.option_value = option_dict["default"]

    def compose(self):
        with Horizontal():
            yield Label(f"{self.option} [gray]{self.option_dict['dest']}[/]")
            with self.prevent(Select.Changed):
                yield Select(
                    value=self.option_dict["default"],
                    options=(
                        (choice, choice) for choice in self.option_dict["choices"]
                    ),
                    allow_blank=False,
                )
        yield Rule()
        return super().compose()

    def on_mount(self):
        self.query_one(Label).tooltip = self.option_dict["help"]
        # Set to current Value
        dest = self.option_dict["dest"]
        if dest != self.option_dict["default"]:
            self.query_one(Select).value = self.app.selected_options_dict[dest]

    def on_select_changed(self, event: Select.Changed):
        self.option_value = event.select.value

    def watch_option_value(self):
        if self.option_value == self.option_dict["default"]:
            self.complete_option = None
        else:
            self.complete_option = f"{self.option}={self.option_value}"

    def watch_was_changed(self):
        if self.was_changed:
            self.query_one(Label).update(
                f"[$success]{self.option}[/] [gray]{self.option_dict['dest']}[/]"
            )
        else:
            self.query_one(Label).update(
                f"{self.option} [gray]{self.option_dict['dest']}[/]"
            )

    def watch_complete_option(self):
        if self.complete_option is None:
            # self.app.options.pop(self.option)
            self.was_changed = False
        else:
            # self.app.options[self.option] = self.complete_option
            self.was_changed = True
        # self.app.update_options()


class ListOption(Vertical):
    app: "AyuApp"
    option: reactive[str] = reactive("", init=False)
    option_value: reactive[list[str]] = reactive(list, init=False)
    complete_option: reactive[str | None] = reactive(None, init=False)
    was_changed: reactive[bool] = reactive(False, init=False)

    def __init__(self, option_dict: dict[str, list], *args, **kwargs) -> None:
        self.option_dict = option_dict
        super().__init__(*args, **kwargs)
        self.classes = "optionwidget"
        self.option = " ".join(option_dict["names"])
        self.option_value = option_dict["default"]

    def on_mount(self):
        self.query_one(Label).tooltip = self.option_dict["help"]
        dest = self.option_dict["dest"]
        dest_values = self.app.selected_options_dict[dest]
        if dest_values != self.option_dict["default"]:
            for value in dest_values:
                self.add_new_value(new_value=value)

    def compose(self):
        with Horizontal():
            yield Label(f"{self.option} [gray]{self.option_dict['dest']}[/]")
            yield Input(placeholder="enter a value and press enter to add to list")
        self.list_table = DataTable(cursor_type="row", show_header=False)
        self.list_table.add_column("option", key="option")
        self.list_table.add_column("remove", key="remove")
        self.list_table.display = False

        yield self.list_table
        yield Rule()
        return super().compose()

    # TODO Add Input validation to prevent adding the same
    def on_input_submitted(self, event: Input.Submitted):
        new_value = event.input.value.strip()
        if new_value:
            self.add_new_value(new_value=new_value)
            event.input.clear()

    def add_new_value(self, new_value):
        if new_value in self.list_table.rows.keys():
            return

        self.list_table.add_row(
            new_value,
            ":cross_mark: click to remove",
            key=new_value,
        )
        self.option_value.append(new_value)
        self.mutate_reactive(ListOption.option_value)

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        self.option_value.remove(event.row_key)
        self.list_table.remove_row(event.row_key)
        self.mutate_reactive(ListOption.option_value)
        # TODO When having a long entry and a scrollbar appears
        # The column widths doesnt reset back

    def watch_option_value(self):
        if self.option_value == self.option_dict["default"]:
            self.complete_option = None
            # self.parent.parent.update_amount()
        else:
            if len(self.option_value) == 1:
                self.complete_option = f"{self.option}={self.option_value[0]}"
            else:
                self.complete_option = f'{self.option}="{",".join(self.option_value)}"'

    def watch_was_changed(self):
        if self.was_changed:
            self.query_one(Label).update(
                f"[$success]{self.option}[/] [gray]{self.option_dict['dest']}[/]"
            )
        else:
            self.query_one(Label).update(
                f"{self.option} [gray]{self.option_dict['dest']}[/]"
            )
            self.query_one(Input).focus()
        self.list_table.display = self.was_changed

    def watch_complete_option(self):
        if self.complete_option is None:
            # self.app.options.pop(self.option)
            self.was_changed = False
        else:
            # self.app.options[self.option] = self.complete_option
            self.was_changed = True
        # self.app.update_options()



from ayu.utils import get_ayu_websocket_host_port


def test_to_come():
    assert True


def test_port_and_host(test_port, test_host):
    host, port = get_ayu_websocket_host_port()

    assert host == "localhost"
    assert port == 1338



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ayu.app import AyuApp


from textual.reactive import reactive
from textual.widgets import Rule, Button
from textual.message import Message
from textual.containers import Vertical, Horizontal

from ayu.utils import TestOutcome


class ToggleRule(Rule):
    test_result: reactive[TestOutcome | None] = reactive(None)
    widget_is_displayed: reactive[bool] = reactive(True)

    class Toggled(Message):
        def __init__(self, togglerule: ToggleRule) -> None:
            self.togglerule: ToggleRule = togglerule
            super().__init__()

        @property
        def control(self) -> ToggleRule:
            return self.togglerule

    def __init__(self, target_widget_id: str, *args, **kwargs) -> None:
        self.target_widget_id = target_widget_id
        super().__init__(*args, **kwargs)

    def compose(self):
        yield Button("[green]Passed[/]")

    def on_button_pressed(self):
        self.widget_is_displayed = not self.widget_is_displayed
        self.post_message(self.Toggled(self))
        self.watch_test_result()

    def watch_test_result(self):
        if self.widget_is_displayed:
            hint_string = " (click to collaps)"
        else:
            hint_string = " (click to open)"

        match self.test_result:
            case TestOutcome.PASSED:
                color = "green"
                result_string = self.test_result
            case TestOutcome.FAILED:
                color = "red"
                result_string = self.test_result
            case TestOutcome.SKIPPED:
                color = "yellow"
                result_string = self.test_result
            case _:
                color = "white"
                hint_string = ""
                result_string = "Please run or select a test"

        self.query_one(
            Button
        ).label = f"[{color}]{result_string}[/][white]{hint_string}[/]"


class ButtonPanel(Vertical):
    app: "AyuApp"
    tests_running: reactive[bool] = reactive(False, init=False)
    file_watcher: reactive[bool] = reactive(False, init=False)

    def on_mount(self):
        # Add Tooltips
        path_to_watch = self.app.test_path or "tests"
        self.query_one(
            "#button_watcher", Button
        ).tooltip = f"If [$success]On[/], tracks file changes under 📁[$warning]{path_to_watch}[/] and reruns tests in changed files automatically"
        self.query_one(
            "#button_plugins", Button
        ).tooltip = "Explore the current plugin options [$warning](changes have no influence on the test execution yet)[/]"
        self.query_one(
            "#button_coverage", Button
        ).tooltip = "Show the current test coverage"

    def compose(self):
        yield Button(label="Plugins", id="button_plugins", variant="warning")
        yield Button(label="Show Log", id="button_log", variant="primary")
        yield Button(label="Show Coverage", id="button_coverage", variant="warning")
        yield Button(label="File Watcher: Off", id="button_watcher", variant="warning")
        with Horizontal():
            yield Button(label="Run tests", id="button_run", variant="success")
            yield Button(label="Cancel Run", id="button_cancel", variant="error")

    def watch_tests_running(self):
        self.query_one("#button_run", Button).disabled = self.tests_running
        self.query_one("#button_cancel", Button).disabled = not self.tests_running

    def watch_file_watcher(self):
        if self.file_watcher:
            self.query_one("#button_watcher", Button).variant = "success"
            self.query_one("#button_watcher", Button).label = "File Watcher: On"
        else:
            self.query_one("#button_watcher", Button).variant = "error"
            self.query_one("#button_watcher", Button).label = "File Watcher: Off"



import os
from pathlib import Path
from textual import work, on
from textual.app import App
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.reactive import reactive
from textual.events import Key
from textual.widgets import Log, Header, Footer, Collapsible, Tree, Button
from textual.containers import Horizontal, Vertical
from textual_tags import Tag
from watchfiles import awatch, PythonFilter

from ayu.event_dispatcher import EventDispatcher
from ayu.constants import WEB_SOCKET_HOST, WEB_SOCKET_PORT
from ayu.utils import (
    EventType,
    NodeType,
    run_all_tests,
    remove_ansi_escapes,
    run_plugin_collection,
    run_test_collection,
)
from ayu.widgets.navigation import TestTree
from ayu.widgets.detail_viewer import DetailView, TestResultDetails
from ayu.widgets.filter import TreeFilter, MarkersFilter
from ayu.widgets.helper_widgets import ToggleRule, ButtonPanel
from ayu.widgets.modals.search import ModalSearch
from ayu.widgets.modals.plugin_manager import ModalPlugin
from ayu.widgets.coverage_explorer import CoverageExplorer
from ayu.widgets.log import OutputLog, LogViewer
from ayu.command_builder import build_command


class AyuApp(App):
    CSS_PATH = Path("assets/ayu.tcss")
    TOOLTIP_DELAY = 0.5

    BINDINGS = [
        Binding("ctrl+l", "run_tests", "Run Tests", show=True, priority=True),
        Binding("ctrl+l", "run_marked_tests", "Run ⭐ Tests", show=True, priority=True),
        Binding("s", "show_details", "Details", show=True),
        Binding("c", "clear_test_results", "Clear Results", show=True, priority=True),
        Binding("ctrl+r", "refresh", "Refresh", show=True, priority=True),
        Binding("O", "open_search", "Search", show=True, priority=True),
        Binding("L", "open_log", "Log", show=True),
        Binding("C", "open_coverage", "Coverage", show=True),
        Binding("P", "open_plugin", "Plugin", show=False),
    ]

    data_test_tree: reactive[dict] = reactive({}, init=False)
    counter_total_tests: reactive[int] = reactive(0, init=False)
    plugin_option_dict: reactive[dict] = reactive({}, init=False)
    selected_options_dict: reactive[dict] = reactive({}, init=False)

    filter: reactive[dict] = reactive(
        {
            "show_favourites": True,
            "show_failed": True,
            "show_skipped": True,
            "show_passed": True,
            "excluded_markers": {},
        },
        init=False,
    )
    test_results_ready: reactive[bool] = reactive(False, init=False)
    tests_running: reactive[bool] = reactive(False, init=False)
    file_watcher: reactive[bool] = reactive(False, init=False)
    markers: reactive[list[str]] = reactive(list)
    DEV: bool = False
    """Show The Log Collapsibles"""

    def __init__(
        self,
        test_path: Path | None = None,
        host: str | None = None,
        port: int | None = None,
        *args,
        **kwargs,
    ):
        self.host = host or os.environ.get("AYU_HOST") or WEB_SOCKET_HOST
        self.port = port or int(os.environ.get("AYU_PORT", 0)) or WEB_SOCKET_PORT
        self.dispatcher = None
        self.test_path = test_path
        super().__init__(*args, **kwargs)

    def compose(self):
        yield Header()
        yield Footer(show_command_palette=False)
        outcome_log = Log(id="log_outcome")
        outcome_log.border_title = "Outcome"
        report_log = Log(id="log_report")
        report_log.border_title = "Report"
        collection_log = Log(id="log_collection")
        collection_log.border_title = "Collection"
        debug_log = Log(id="log_debug")
        debug_log.border_title = "Debug"
        yield LogViewer()
        yield CoverageExplorer()
        with Horizontal():
            with Vertical(id="vertical_test_tree"):
                yield TestTree(label="Tests", id="testtree").data_bind(
                    filter=AyuApp.filter,
                    filtered_data_test_tree=AyuApp.data_test_tree,
                    filtered_counter_total_tests=AyuApp.counter_total_tests,
                )
                yield TreeFilter().data_bind(
                    test_results_ready=AyuApp.test_results_ready, markers=AyuApp.markers
                )
            with Vertical():
                yield DetailView()
                if self.DEV:
                    with Collapsible(title="Outcome", collapsed=True):
                        yield outcome_log
                    with Collapsible(title="Report", collapsed=True):
                        yield report_log
                    with Collapsible(title="Collection", collapsed=True):
                        yield collection_log
                    with Collapsible(title="Debug", collapsed=False):
                        yield debug_log
                yield ButtonPanel().data_bind(
                    tests_running=AyuApp.tests_running, file_watcher=AyuApp.file_watcher
                )

    async def on_load(self):
        self.start_socket()

        # Watcher
        # self.run_tests_on_change()

    @work(
        exclusive=True,
        description="Watches files for changes to automatically rerun the changed file",
        group="Watcher",
    )
    async def start_tests_on_change_worker(self):
        path_to_watch = self.test_path or Path("tests")
        async for changes in awatch(path_to_watch, watch_filter=PythonFilter()):
            if not self.tests_running:
                self.notify("Files have changed", severity="information")
                for change in changes:
                    change_mode, path = change
                    self.action_run_tests(tests_to_run=Path(path))
            else:
                self.notify(
                    title="Warning", message="Tests running already", severity="warning"
                )

    def on_mount(self):
        # For Developing/Debugging
        if self.DEV:
            self.dispatcher.register_handler(
                event_type=EventType.OUTCOME,
                handler=lambda msg: self.update_outcome_log(msg),
            )
            self.dispatcher.register_handler(
                event_type=EventType.COVERAGE,
                handler=lambda msg: self.update_debug_log(msg),
            )
            self.dispatcher.register_handler(
                event_type=EventType.DEBUG,
                handler=lambda msg: self.update_debug_log(msg),
            )
            self.dispatcher.register_handler(
                event_type=EventType.REPORT,
                handler=lambda msg: self.update_report_log(msg),
            )

        self.dispatcher.register_handler(
            event_type=EventType.PLUGIN,
            handler=lambda msg: self.update_plugin_dict(msg),
        )
        self.dispatcher.register_handler(
            event_type=EventType.OPTIONS,
            handler=lambda msg: self.update_selected_options(msg),
        )

        self.dispatcher.register_handler(
            event_type=EventType.COLLECTION,
            handler=lambda data: self.update_app_data(data),
        )
        self.query_one(TestTree).focus()

        self.collect_initial_plugins()
        self.collect_initial_test_tree()

    def update_app_data(self, data):
        self.data_test_tree = data["tree"]
        self.counter_total_tests = data["meta"]["test_count"]
        self.markers = data["meta"]["markers"]

    def update_plugin_dict(self, data):
        # if not self.plugin_dict:
        self.plugin_option_dict = data["plugin_dict"]
        # self.notify(f"{self.plugin_option_dict.keys()}", markup=False)

    def update_selected_options(self, data):
        # if not self.plugin_dict:
        # if not self.selected_options_dict:
        self.selected_options_dict.update(data["option_dict"])

    # Get initial Data
    @work(
        exclusive=True,
        group="Collector",
        description="Executes `pytest --co` and collects tests",
    )
    async def collect_initial_test_tree(self):
        command = build_command(
            plugins=None,
            tests_to_run=self.test_path,
            pytest_options=["--co"],
        )
        await run_test_collection(command=command)

    @work(
        exclusive=True,
        group="Plugin",
        description="Executes `pytest --help` and collect Plugin Infos",
    )
    async def collect_initial_plugins(self):
        command = build_command(
            plugins=None,
            tests_to_run=self.test_path,
            pytest_options=["--help"],
        )
        await run_plugin_collection(command=command)

    @work(exclusive=True, description="Keeps the websocket alive", group="Websocket")
    async def start_socket(self):
        self.dispatcher = EventDispatcher(host=self.host, port=self.port)
        self.notify(
            f"Websocket Started at\n[orange]{self.host}:{self.port}[/]", timeout=1
        )
        try:
            await self.dispatcher.start()
        except OSError as e:
            self.log.error(e)
            pass

    def on_key(self, event: Key):
        if event.key == "w":
            for worker in self.workers:
                self.notify(f"{worker}", markup=False)

    @on(Button.Pressed, ".filter-button")
    def update_test_tree_filter(self, event: Button.Pressed):
        button_id_part = event.button.id.split("_")[-1]
        filter_state = event.button.filter_is_active
        self.filter[f"show_{button_id_part}"] = filter_state
        self.mutate_reactive(AyuApp.filter)

    def reset_filters(self):
        for btn in self.query(".filter-button"):
            btn.filter_is_active = True
        self.filter = {
            "show_favourites": True,
            "show_failed": True,
            "show_skipped": True,
            "show_passed": True,
        }
        self.mutate_reactive(AyuApp.filter)

    @on(Tag.Hovered)
    @on(Tag.Focused)
    @on(Tag.Selected)
    def hightlight_test_tree(self, event: Tag.Hovered | Tag.Focused | Tag.Selected):
        self.query_one(TestTree).highlight_marker_rows(marker=event.tag.value)

    @on(MarkersFilter.Marked)
    def favourite_tests_from_tags(self, event: MarkersFilter.Marked):
        self.query_one(TestTree).mark_test_as_fav_from_markers(marker=event.current_tag)

    @on(ModalSearch.Marked)
    def favourite_tests_from_search(self, event: ModalSearch.Marked):
        self.query_one(TestTree).action_mark_test_as_fav_from_search(
            nodeid=event.nodeid
        )

    @on(Tree.NodeHighlighted)
    def update_test_preview(self, event: Tree.NodeHighlighted):
        detail_view = self.query_one(DetailView)
        detail_view.file_path_to_preview = Path(event.node.data["path"])
        if event.node.data["type"] in [
            NodeType.FUNCTION,
            NodeType.COROUTINE,
            NodeType.CLASS,
        ]:
            detail_view.test_start_line_no = event.node.data["lineno"]
        else:
            detail_view.test_start_line_no = -1

        self.query_one(ToggleRule).test_result = event.node.data["status"]
        self.query_one(TestResultDetails).selected_node_id = event.node.data["nodeid"]

    @on(Button.Pressed, "#button_watcher")
    def toggle_file_watcher(self, event: Button.Pressed):
        self.file_watcher = not self.file_watcher

    @on(Button.Pressed, "#button_plugins")
    def open_plugin_screen(self, event: Button.Pressed):
        self.action_open_plugin()

    @on(Button.Pressed, "#button_coverage")
    def toggle_coverage_explorer(self, event: Button.Pressed):
        self.action_open_coverage()

    @on(Button.Pressed, "#button_log")
    def toggle_log_viewer(self, event: Button.Pressed):
        self.action_open_log()

    @on(Button.Pressed, "#button_run")
    def toggle_test_run(self, event: Button.Pressed):
        if self.query_one(TestTree).marked_tests:
            self.action_run_marked_tests()
        else:
            self.action_run_tests()

    # Actions
    def action_open_plugin(self):
        self.push_screen(
            ModalPlugin().data_bind(
                selected_options_dict=AyuApp.selected_options_dict,
                plugin_option_dict=AyuApp.plugin_option_dict,
            )
        )

    def action_show_details(self):
        self.query_one(DetailView).toggle()
        self.query_one(TreeFilter).toggle()

    def action_open_log(self):
        self.query_one(LogViewer).display = not self.query_one(LogViewer).display

    def action_open_coverage(self):
        cov_explorer = self.query_one(CoverageExplorer)
        cov_explorer.disabled = cov_explorer.display
        cov_explorer.display = not cov_explorer.display
        if cov_explorer.display:
            cov_explorer.query_one("#table_coverage").focus()
        else:
            self.query_one(TestTree).focus()

    @work(
        thread=True,
        group="Test Runner",
        description="runs all tests under a specific path",
    )
    async def action_run_tests(self, tests_to_run: Path | None = None):
        self.tests_running = True
        self.reset_filters()
        # Log Runner Output
        if not tests_to_run:
            tests_to_run = self.test_path

        command = build_command(
            plugins=None,
            tests_to_run=tests_to_run,
        )
        runner = await run_all_tests(command=command)
        while runner:
            if runner.returncode is not None:
                break
            output_line = await runner.stdout.readline()
            decoded_line = remove_ansi_escapes(output_line.decode())
            self.call_from_thread(self.query_one(OutputLog).write_line, decoded_line)
        # Log Runner End
        self.tests_running = False
        self.test_results_ready = True

    @work(thread=True, group="Test Runner", description="runs marked tests only")
    async def action_run_marked_tests(self):
        self.tests_running = True
        self.reset_filters()

        command = build_command(
            plugins=None,
            tests_to_run=self.query_one(TestTree).marked_tests,
        )
        runner = await run_all_tests(command=command)
        while runner:
            if runner.returncode is not None:
                break
            output_line = await runner.stdout.readline()
            decoded_line = remove_ansi_escapes(output_line.decode())
            self.call_from_thread(self.query_one(OutputLog).write_line, decoded_line)

        self.tests_running = False
        self.test_results_ready = True

    def action_refresh(self):
        self.collect_initial_test_tree()
        self.test_results_ready = False

    def action_clear_test_results(self):
        self.test_results_ready = False
        self.query_one(TestTree).reset_test_results()
        for log in self.query(Log):
            log.clear()

    def action_open_search(self):
        def select_searched_nodeid(nodeid: str | None):
            if nodeid:
                node = self.query_one(TestTree).get_node_by_nodeid(nodeid=nodeid)
                self.query_one(TestTree).select_node(node=node)

        self.push_screen(ModalSearch(), callback=select_searched_nodeid)

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        # on app startup widget is not mounted yet so
        # try except is needed
        try:
            if action == "run_tests":
                if self.query_one(TestTree).marked_tests:
                    return False
            if action == "run_marked_tests":
                if not self.query_one(TestTree).marked_tests:
                    return False
        except NoMatches:
            return True
        return True

    def update_outcome_log(self, msg):
        self.query_one("#log_outcome", Log).write_line(f"{msg}")

    def update_report_log(self, msg):
        self.query_one("#log_report", Log).write_line(f"{msg}")

    def update_debug_log(self, msg):
        self.query_one("#log_debug", Log).write_line(f"{msg}")

    def watch_file_watcher(self):
        if self.file_watcher:
            self.start_tests_on_change_worker()
        else:
            self.workers.cancel_group(node=self, group="Watcher")

    def watch_tests_running(self):
        if self.tests_running:
            self.notify("Started Tests", timeout=2)

    # Watchers
    # def watch_data_test_tree(self):
    #     self.query_one("#log_collection", Log).write_line(f"{self.data_test_tree}")


# https://watchfiles.helpmanual.io



from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ayu.app import AyuApp

from textual import on
from textual.reactive import reactive
from textual.widgets import TextArea
from textual_slidecontainer import SlideContainer
from rich.text import Text

from ayu.utils import EventType, get_preview_test
from ayu.widgets.helper_widgets import ToggleRule


class DetailView(SlideContainer):
    file_path_to_preview: reactive[Path | None] = reactive(None, init=False)
    test_start_line_no: reactive[int] = reactive(-1, init=False)

    def __init__(self, *args, **kwargs):
        super().__init__(
            slide_direction="up",
            floating=False,
            start_open=False,
            duration=0.5,
            *args,
            **kwargs,
        )

    def compose(self):
        yield CodePreview("Please select a test", id="textarea_preview")
        yield ToggleRule(target_widget_id="textarea_test_result_details")
        yield TestResultDetails(id="textarea_test_result_details")

    def watch_file_path_to_preview(self):
        self.border_title = self.file_path_to_preview.as_posix()

    def watch_test_start_line_no(self):
        if self.test_start_line_no == -1:
            self.query_one("#textarea_preview", TextArea).text = "Please select a test"
        else:
            content = get_preview_test(
                file_path=self.file_path_to_preview,
                start_line_no=self.test_start_line_no,
            )

            self.query_one(
                "#textarea_preview", TextArea
            ).line_number_start = self.test_start_line_no
            self.query_one("#textarea_preview", TextArea).text = content

    @on(ToggleRule.Toggled)
    def toggle_code_result_visibility(self, event: ToggleRule.Toggled):
        target_widget = self.query_one(f"#{event.togglerule.target_widget_id}")
        target_widget.display = not target_widget.display


class CodePreview(TextArea):
    def on_mount(self):
        self.language = "python"
        self.read_only = True
        self.show_line_numbers = True


class TestResultDetails(TextArea):
    app: "AyuApp"
    selected_node_id: reactive[str] = reactive("")
    report_data: reactive[dict] = reactive({})

    def on_mount(self):
        # self.language = "python"
        self.read_only = True
        # self.styles.color = 'red'

        self.app.dispatcher.register_handler(
            event_type=EventType.REPORT,
            handler=lambda msg: self.update_report_data(msg),
        )

    def update_report_data(self, data: dict):
        if data["report"]:
            self.report_data = data["report"]

    def watch_selected_node_id(self):
        if self.report_data.get(self.selected_node_id):
            text = self.report_data[self.selected_node_id]["longreprtext"]
            self.text = self.make_error_part_red(text)

            duration = self.report_data[self.selected_node_id]["duration"]
            self.parent.border_subtitle = Text.from_markup(
                f":hourglass_not_done: {duration:.5f}s"
            )

            self.scroll_end(animate=False)
        else:
            self.text = ""
            self.parent.border_subtitle = ""

    def make_error_part_red(self, text: str) -> str:
        """highlight lines which caused the error"""
        # TODO
        return text



import sys
import os

import pytest

from ayu.app import AyuApp


# @pytest.mark.xdist_group(name="group1")
@pytest.mark.skipif(sys.platform.startswith("win"), reason="Windows is too slow")
async def test_app_screen(testcase_path):
    os.environ["AYU_PORT"] = "1338"
    os.environ["AYU_HOST"] = "localhost"

    test_app = AyuApp(test_path=testcase_path)
    async with test_app.run_test() as pilot:
        # Wait for test collection
        await pilot.pause(3)

        assert pilot.app.data_test_tree


# @pytest.mark.xdist_group(name='group1')
# class Test_App:
#
#     os.environ["AYU_PORT"] = "1338"
#     os.environ["AYU_HOST"] = "localhost"
#
#     @pytest.fixture(autouse=True)
#     def test_app(self, testcase_path):
#         self.app = AyuApp(test_path=testcase_path)
#
#     async def test_app_vars(self):
#         async with self.app.run_test() as pilot:
#             assert pilot.app.test_path.as_posix() == "tests/test_cases"
#             assert not pilot.app.test_results_ready
#             assert not pilot.app.tests_running
#
#     async def test_app_ports(self, test_port, test_host):
#         async with self.app.run_test() as pilot:
#             assert pilot.app.host == test_host
#             assert pilot.app.port == test_port
#
#     @pytest.mark.skipif(sys.platform.startswith("win"), reason="Windows is too slow")
#     async def test_app_screen(self):
#         async with self.app.run_test() as pilot:
#             # Wait for test collection
#             pilot.app.action_refresh()
#             await pilot.pause(2)
#
#             assert pilot.app.data_test_tree



from textual.containers import Center
from textual.widgets import Log, Label


class LogViewer(Center):
    def on_mount(self):
        self.display = False

    def compose(self):
        yield LogLabel("[bold]Log Output[/]")
        yield OutputLog(highlight=True)


class LogLabel(Label):
    "Label for log"


class OutputLog(Log):
    can_focus = False
    """Log to display command outputs"""



from pathlib import Path
from dataclasses import dataclass

from ayu.utils import ayu_is_run_as_tool


@dataclass
class Plugin:
    name: str
    is_installed: bool
    options: list[str]
    version: str = ""


def build_command(
    plugins: list[Plugin] | None = None,
    tests_to_run: Path | list[str] | None = None,
    pytest_options: list[str] | None = None,
) -> str:
    command_parts = []

    # is ayu installed in the current project or run as tool
    is_tool = ayu_is_run_as_tool()

    # install ayu on the fly, if ayu is not installed in the project itself
    substring_uv = "uv run --with ayu" if is_tool else "uv run"
    command_parts.append(substring_uv)

    # if plugins are declared with options, but are not installed in the project install them on the fly
    if plugins:
        substring_plugins_to_install = " ".join(
            [f"--with {plugin.name}" for plugin in plugins if not plugin.is_installed]
        )
        command_parts.append(substring_plugins_to_install)

    # pytest call here
    command_parts.append("pytest")

    # plugin options
    if plugins:
        substring_plugin_options = " ".join(
            [option for plugin in plugins for option in plugin.options]
        )
        command_parts.append(substring_plugin_options)

    # plugin options
    if pytest_options:
        substring_pytest_options = " ".join(pytest_options)
        command_parts.append(substring_pytest_options)
    # wrap tests in quotes to prevent errors when parametrize
    # args include spaces
    if tests_to_run:
        if isinstance(tests_to_run, Path):
            substring_tests_to_run = tests_to_run.as_posix()
        else:
            substring_tests_to_run = " ".join([f'"{test}"' for test in tests_to_run])
        command_parts.append(substring_tests_to_run)

    # only join non empty substrings
    return " ".join([substring for substring in command_parts if substring])



from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from ayu.app import AyuApp
from textual.message import Message
from textual.events import Key
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.binding import Binding
from textual.widgets import Input, Footer, Label
from textual.content import Content
from textual.containers import Center

from textual_autocomplete import AutoComplete, DropdownItem, TargetState

from ayu.utils import NodeType


class SearchInput(Input):
    filtered_node_types: reactive[list] = reactive([])

    def on_key(self, event: Key):
        if event.key == "backspace":
            if not self.value and self.filtered_node_types:
                self.filtered_node_types.pop()
                self.mutate_reactive(SearchInput.filtered_node_types)
                # Fix error showing latest autocompletes if backspacing from 2 elements

    def watch_filtered_node_types(self):
        if self.filtered_node_types:
            hint_substring = (
                f"(press `backspace` to remove {self.filtered_node_types[-1]})"
            )
            self.placeholder = f"Search only for tests of type {', '.join(self.filtered_node_types)} {hint_substring}"
            self.styles.border_bottom = ("tall", self.styles.background)
            self.border_subtitle = Content.from_markup(
                "[white]active filters: [/]"
                + f"[$success-darken-2]{', '.join(self.filtered_node_types)}[/]"
            )
        else:
            self.placeholder = (
                "Type to search for test or press ':' to filter for different NodeTypes"
            )
            self.styles.border_bottom = None


class SearchAutoComplete(AutoComplete):
    app: "AyuApp"

    def get_candidates(self, target_state: TargetState) -> list[DropdownItem]:
        # Filter candidates based on target_state.text
        nodes = self.app.query_one("#testtree").test_nodes
        prefix_bg = "$surface-lighten-3"
        if target_state.text.startswith(":"):
            return [
                DropdownItem(
                    main=Content.from_markup(
                        f"[{prefix_bg} on {prefix_bg}]:[/][on {prefix_bg}]{node_type.value}[/][{prefix_bg}]\ue0b4[/]"
                    )
                )
                for node_type in NodeType
                if node_type not in self.target.filtered_node_types
            ]
        return [
            DropdownItem(
                main=f"{node.data['nodeid']}",
                prefix=Content.from_markup(
                    f"[on {prefix_bg}] {node.data['type']}[/][{prefix_bg}]\ue0b4[/] {'⭐' if node.data['favourite'] else ''}"
                ),
            )
            for node in nodes[1:]
            if node.data["type"] in (self.target.filtered_node_types or NodeType)
            # if node.data['type'] in
        ]

    def get_search_string(self, target_state: TargetState) -> str:
        # get only part after certain filter
        if target_state.text.startswith(":"):
            return target_state.text
        return target_state.text[: target_state.cursor_position]

    # Add Filter choice
    def post_completion(self) -> None:
        value = self.target.value
        if value[1:-1] in NodeType:
            self.target.filtered_node_types.append(self.target.value[1:-1])
            self.target.mutate_reactive(SearchInput.filtered_node_types)
            self.target.clear()
            self.set_timer(0.05, self.action_show)
        self.action_hide()

    # Override to prevent aligning with input cursor
    def _align_to_target(self) -> None:
        return

    def action_show(self) -> None:
        super().action_show()
        self.refresh_bindings()

    def action_hide(self) -> None:
        self.refresh_bindings()
        return super().action_hide()


class ModalSearch(ModalScreen):
    app: "AyuApp"

    BINDINGS = [
        Binding("escape", "app.pop_screen", "Close", show=True),
        Binding(
            "ctrl+j", "navigate_highlight('down')", "down", priority=True, show=True
        ),
        Binding("ctrl+k", "navigate_highlight('up')", "up", priority=True, show=True),
        Binding("ctrl+f", "mark_as_fav", "⭐ Mark", priority=True, show=True),
    ]

    class Marked(Message):
        def __init__(self, nodeid: str) -> None:
            self.nodeid = nodeid
            super().__init__()

        @property
        def control(self):
            return self.nodeid

    def compose(self):
        with Center():
            yield Label("Search for tests")
            yield SearchInput(id="input_search")
            yield Footer()
            yield SearchAutoComplete(
                target="#input_search",
            )

    def action_navigate_highlight(self, direction: Literal["up", "down"]):
        """go to next hightlight in completion option list"""
        if not isinstance(self.app.focused, Input):
            return
        option_list = self.query_one(SearchAutoComplete).option_list
        displayed = self.query_one(SearchAutoComplete).display
        highlighted = option_list.highlighted
        int_direction = 1 if direction == "down" else -1

        if displayed:
            highlighted = (highlighted + int_direction) % option_list.option_count
        else:
            self.query_one(SearchAutoComplete).action_show()
            highlighted = 0

        option_list.highlighted = highlighted

    def get_last_state(self, highlighted, scroll_y):
        autocomplete = self.query_one(SearchAutoComplete)
        target_state = autocomplete._get_target_state()

        search_string = autocomplete.get_search_string(target_state)
        autocomplete._rebuild_options(target_state, search_string)

        autocomplete.option_list.highlighted = highlighted
        autocomplete.option_list.scroll_y = scroll_y

    def action_mark_as_fav(self):
        option_list = self.query_one(SearchAutoComplete).option_list
        # displayed = self.query_one(SearchAutoComplete).display

        # Get current hightlight and Scrollposition
        highlighted = option_list.highlighted
        scroll_y = option_list.scroll_y

        node_to_mark = option_list.options[highlighted].value
        self.post_message(self.Marked(nodeid=node_to_mark))

        # Get rebuild options to display changes
        self.set_timer(
            delay=0.05,
            callback=lambda: self.get_last_state(
                highlighted=highlighted, scroll_y=scroll_y
            ),
        )

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action == "mark_as_fav":
            return self.query_one(SearchAutoComplete).display and not self.query_one(
                SearchInput
            ).value.startswith(":")
        return True

    def on_input_submitted(self, event: Input.Submitted):
        self.dismiss(result=self.query_one(SearchInput).value)



import sys
import os

import pytest

from ayu.app import AyuApp
from ayu.widgets.modals.plugin_manager import ModalPlugin


# @pytest.mark.xdist_group(name="group2")
@pytest.mark.skipif(sys.platform.startswith("win"), reason="Windows is too slow")
async def test_app_plugins(testcase_path):
    os.environ["AYU_PORT"] = "1339"
    os.environ["AYU_HOST"] = "localhost"
    test_app = AyuApp(test_path=testcase_path)
    async with test_app.run_test() as pilot:
        # Wait for test collection
        await pilot.press("P")
        await pilot.pause(5)
        assert isinstance(pilot.app.screen, ModalPlugin)

        for plugin in ["ayu", "asyncio", "xdist", "pytest_cov"]:
            assert plugin in list(pilot.app.plugin_option_dict.keys())



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ayu.app import AyuApp

from textual import on, work
from textual.binding import Binding
from textual.reactive import reactive
from textual.containers import Horizontal, Vertical
from textual.widgets import Label, DataTable, TextArea

from ayu.utils import EventType, build_bar


class CoverageExplorer(Vertical):
    app: "AyuApp"

    coverage_dict: reactive[dict] = reactive({})
    selected_file: reactive[str] = reactive("")
    selected_line: reactive[list] = reactive([])

    def on_mount(self):
        self.display = False

        self.app.dispatcher.register_handler(
            event_type=EventType.COVERAGE,
            handler=lambda msg: self.update_coverage_dict(msg),
        )

    def compose(self):
        yield CoverageLabel("[bold]Test Coverage[/]")
        with Vertical():
            with Horizontal():
                yield CoverageTable(id="table_coverage").data_bind(
                    CoverageExplorer.coverage_dict
                )
                yield MissingLinesTable(id="table_lines").data_bind(
                    coverage_dict=CoverageExplorer.coverage_dict,
                    selected_file=CoverageExplorer.selected_file,
                )
            yield CoverageFilePreview("Preview").data_bind(
                selected_file=CoverageExplorer.selected_file,
                selected_line=CoverageExplorer.selected_line,
            )

    def watch_disabled(self, disabled: bool) -> None:
        return super().watch_disabled(disabled)

    def update_coverage_dict(self, msg):
        self.coverage_dict = msg["coverage_dict"]

    @on(DataTable.RowHighlighted, "#table_coverage")
    def update_selected_file(self, event: DataTable.RowHighlighted):
        if event.row_key:
            file_name = self.query_one(CoverageTable).get_row(event.row_key)[0]
            self.selected_file = file_name

    @on(DataTable.RowHighlighted, "#table_lines")
    # use an exclusive worker here to support holding down j/k
    # which would otherwise give a IndexError
    @work(thread=True, exclusive=True, description="Update Selected Line")
    def update_selected_line(self, event: DataTable.RowHighlighted):
        if event.row_key:
            if isinstance(self.app.focused, CoverageTable):
                self.selected_line = self.coverage_dict[self.selected_file][
                    "lines_missing"
                ][0]
            else:
                line = self.query_one(MissingLinesTable).get_row(event.row_key)[0]
                self.selected_line = line


class CoverageLabel(Label):
    "Label for Coverage Explorer"


class CoverageTable(DataTable):
    """General Table for Coverage Information"""

    BINDINGS = [
        Binding("j, down", "cursor_down", "down", key_display="j/↓"),
        Binding("k, up", "cursor_up", "up", key_display="k/↑"),
        Binding("l, right", "go_to_lines", "to lines"),
    ]

    coverage_dict: reactive[dict] = reactive({})
    COLUMNS = ["Name", "Statements", "Missing", "Covered", ""]

    def on_mount(self):
        self.cursor_type = "row"
        self.border_title = "Coverage Report"

        for column in self.COLUMNS:
            self.add_column(label=column, width=None if column else 10)

    # Go to first row, when navigating down on last row
    def action_cursor_down(self) -> None:
        if self.cursor_coordinate.row == (self.row_count - 1):
            self.move_cursor(row=0)
            return
        return super().action_cursor_down()

    # Go to last row, when navigating up on first row
    def action_cursor_up(self) -> None:
        if self.cursor_coordinate.row == 0:
            self.move_cursor(row=self.row_count - 1)
            return
        return super().action_cursor_up()

    def watch_coverage_dict(self):
        if not self.coverage_dict:
            return

        current_line = self.cursor_row or 0
        self.clear()

        for module_name, module_dict in self.coverage_dict.items():
            self.add_row(
                module_name,
                module_dict["n_statements"],
                module_dict["n_missed"],
                f"{module_dict['percent_covered']:05.2f}%",
                build_bar(module_dict["percent_covered"]),
                key=module_name,
            )

        # go to last known cursor position
        self.move_cursor(row=current_line)

    @on(DataTable.RowSelected)
    def test(self, event: DataTable.RowSelected): ...

    def action_go_to_lines(self):
        self.app.action_focus_next()


class MissingLinesTable(DataTable):
    """Table for Missing Lines"""

    BINDINGS = [
        Binding("j, down", "cursor_down", "down", key_display="j/↓"),
        Binding("k, up", "cursor_up", "up", key_display="k/↑"),
        Binding("h, left", "back_to_coverage", "to cov table"),
    ]
    coverage_dict: reactive[dict] = reactive({})
    selected_file: reactive[str] = reactive("")

    def on_mount(self):
        self.cursor_type = "row"
        self.add_columns(
            "Missing Lines",
        )

    # Go to first row, when navigating down on last row
    def action_cursor_down(self) -> None:
        if self.cursor_coordinate.row == (self.row_count - 1):
            self.move_cursor(row=0)
            return
        return super().action_cursor_down()

    # Go to last row, when navigating up on first row
    def action_cursor_up(self) -> None:
        if self.cursor_coordinate.row == 0:
            self.move_cursor(row=self.row_count - 1)
            return
        return super().action_cursor_up()

    def watch_selected_file(self):
        if not self.selected_file:
            return

        self.clear()

        for missing_lines in self.coverage_dict[self.selected_file]["lines_missing"]:
            self.add_row(missing_lines)

    def action_back_to_coverage(self):
        self.app.action_focus_previous()


class CoverageFilePreview(TextArea):
    """Preview of Lines in source file"""

    selected_file: reactive[str] = reactive("")
    selected_line: reactive[int] = reactive(0)

    def on_mount(self):
        self.language = "python"
        self.show_line_numbers = True
        self.read_only = True
        self.border_title = "File Preview"

    def watch_selected_file(self):
        if self.selected_file:
            with open(self.selected_file, "r") as file:
                self.text = file.read()
            self.border_title = self.selected_file

    def watch_selected_line(self):
        if self.selected_line:
            self.move_cursor(location=(self.selected_line - 1, 0))

            self.scroll_to(
                y=self.selected_line - 5,  # 4 rows space on top
                animate=True,
                duration=0.1,
                force=True,
            )



import os
import asyncio
import pytest
from pytest import Config, TestReport, Session, Parser
from _pytest.terminal import TerminalReporter

from ayu.event_dispatcher import send_event, check_connection
from ayu.classes.event import Event
from ayu.utils import (
    EventType,
    TestOutcome,
    get_pytest_current_options,
    remove_ansi_escapes,
    build_dict_tree,
    build_plugin_dict,
    get_coverage_data,
)

# import logging
# logging.basicConfig(level=logging.DEBUG)


def pytest_addoption(parser: Parser) -> None:
    group = parser.getgroup("ayu", "interactive pytest interface")
    group.addoption(
        "--disable-ayu",
        "--da",
        action="store_true",
        default=False,
        help="Disable Ayu plugin functionality, i.e. do not send events to websocket",
    )


@pytest.hookimpl(trylast=True)
def pytest_configure(config: Config) -> None:
    if not config.getoption("--disable-ayu"):
        config.pluginmanager.register(Ayu(config))


class Ayu:
    def __init__(self, config: Config):
        self.config = config
        self.connected = False

        if check_connection():
            print("Websocket connected")
            self.connected = True
        else:
            self.connected = False
            print("Websocket not connected")
        self.load_current_options()
        self.load_used_plugin_infos()

    def load_current_options(self):
        if self.connected and self.config.getoption("--help"):
            option_dict = get_pytest_current_options(conf=self.config)
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.OPTIONS,
                        event_payload={"option_dict": option_dict},
                    )
                )
            )

    def load_used_plugin_infos(self):
        if self.connected and self.config.getoption("--help"):
            plugin_dict = build_plugin_dict(conf=self.config)
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.PLUGIN,
                        event_payload={"plugin_dict": plugin_dict},
                    )
                )
            )

    # must tryfirst, otherwise collection-only is returning
    @pytest.hookimpl(tryfirst=True)
    def pytest_runtestloop(self, session: Session):
        if self.connected and session.config.getoption("--collect-only"):
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.DEBUG,
                        event_payload={"test": "test"},
                        # event_payload={"no_items":session.testscollected,"items":f"{session.items}"},
                    )
                )
                # ,debug_mode=True
            )

    # build test tree
    def pytest_collection_finish(self, session: Session):
        if self.connected:
            print("Connected to Ayu")
            if session.config.getoption("--collect-only"):
                tree = build_dict_tree(items=session.items)
                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.COLLECTION,
                            event_payload=tree,
                        )
                    ),
                )
            else:
                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.SCHEDULED,
                            event_payload=[item.nodeid for item in session.items],
                        )
                    )
                )
        return

    # gather status updates during run
    def pytest_runtest_logreport(self, report: TestReport):
        if self.config.pluginmanager.hasplugin("xdist") and (
            "PYTEST_XDIST_WORKER" not in os.environ
        ):
            pass
            return

        is_relevant = (report.when == "call") or (
            (report.when == "setup")
            and (report.outcome.upper() in [TestOutcome.FAILED, TestOutcome.SKIPPED])
        )

        if self.connected and is_relevant:
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.OUTCOME,
                        event_payload={
                            "nodeid": report.nodeid,
                            "outcome": report.outcome.upper(),
                        },
                    )
                )
            )

    # summary after run for each tests
    @pytest.hookimpl(trylast=True)
    def pytest_terminal_summary(self, terminalreporter: TerminalReporter):
        # Debugging
        # from pprint import pprint
        # option_dict = {option:value for option, value in self.config.option._get_kwargs()}
        # pprint(option_dict)

        # Summary part of individual Workers
        if self.config.pluginmanager.hasplugin("xdist") and (
            "PYTEST_XDIST_WORKER" not in os.environ
        ):
            # Needs to run within workers for correct report
            # if pytest-xdist is available
            if self.config.pluginmanager.hasplugin("_cov") and self.connected:
                coverage_dict = get_coverage_data()

                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.COVERAGE,
                            event_payload={
                                "coverage_dict": coverage_dict,
                            },
                        )
                    )
                )

        else:
            if self.config.pluginmanager.hasplugin("_cov") and self.connected:
                coverage_dict = get_coverage_data()

                asyncio.run(
                    send_event(
                        event=Event(
                            event_type=EventType.COVERAGE,
                            event_payload={
                                "coverage_dict": coverage_dict,
                            },
                        )
                    )
                )

        report_dict = {}
        # warning report has no report.when
        for outcome, reports in terminalreporter.stats.items():
            # raise Exception(terminalreporter.stats.keys())
            if outcome in ["", "deselected"]:
                continue
            for report in reports:
                report_dict[report.nodeid] = {
                    "nodeid": report.nodeid,
                    # Not in warning report
                    # TODO Handle warning reports
                    "when": report.when,
                    "caplog": report.caplog,
                    "longreprtext": remove_ansi_escapes(report.longreprtext),
                    "duration": report.duration,
                    "outcome": report.outcome,
                    "lineno": report.location[1],
                    "otherloc": report.location[2],
                }

        # import json

        if self.connected:
            asyncio.run(
                send_event(
                    event=Event(
                        event_type=EventType.REPORT,
                        event_payload={
                            "report": report_dict,
                        },
                    )
                )
            )
